# Test various formulas for "elemz"
# using computations with random elements of the rational field.

def random_rational_list(n):
   s = set()
   num_bound = 10 + n
   den_bound = 10 + n
   while len(s) < n:
      q = QQ.random_element(num_bound, den_bound)
      s.add(q)
   return list(s)

def random_rational_list_without_zeros(n):
   s = set()
   num_bound = 10 + n
   den_bound = 10 + n
   while len(s) < n:
      q = QQ.random_element(num_bound, den_bound)
      if q != 0:
         s.add(q)
   return list(s)

def mul_polynomial_by_binomial(acoefs, b):
   alen = len(acoefs)
   midcoefs = [acoefs[j] * b + acoefs[j - 1] for j in range(1, alen)]
   return [b * acoefs[0]] + midcoefs + [acoefs[alen - 1]]

def elem_pols(t):
   m = len(t)
   result = [1]
   for j in range(m):
      result = mul_polynomial_by_binomial(result, t[j])
   return list(reversed(result))

def elemz_via_lincomb_elem(z):
   n = len(z)
   elem_z = elem_pols(z)
   result = [0] * (2 * n + 1)
   for m in range(2 * n + 1):
      m1 = max(0, m - n)
      m2 = floor(m / 2)
      result[m] = sum([binomial(n - m + 2 * k, k) * elem_z[m - 2 * k] for k in range(m1, m2 + 1)])
   return result

def test_elemz_via_lincomb_elem_numer(n, verbos):
   if verbos >= 2:
      print('test_elemz_via_lincomb_elem_numer, n = %d' % n)
   x = random_rational_list_without_zeros(n)
   y = list(x) + [1 / x[j] for j in range(n)]
   z = [x[j] + 1 / x[j] for j in range(n)]
   res0 = elem_pols(y)
   res1 = elemz_via_lincomb_elem(z)
   if verbos >= 3:
      print('res0 = ' + str(res0))
      print('res1 = ' + str(res1))
   return res0 == res1

def big_test_elemz_via_lincomb_elem_numer(nmax, nrep, verbos):
   if verbos >= 1:
      print('big_test_elemz_via_lincomb_elem_numer, nmax = %d, nrep = %d' % (nmax, nrep))
   tasks = [(n, r) for n in range(1, nmax + 1) for r in range(nrep)]
   return all(test_elemz_via_lincomb_elem_numer(n, verbos) for (n, r) in tasks)

def prod_difs(x, j):
   return prod(x[j] - x[k] for k in range(j) + range(j + 1, len(x)))

def vandermonde(x):
   n = len(x)
   return prod([prod_difs(x[j : ], 0) for j in range(n - 1)])

def tvandermonde(x, la):
   n = len(x)
   F = x[0].parent()
   a = matrix(F, n, n)
   for j in range(n):
      for k in range(n):
         d = la[j] + n - 1 - j
         a[j, k] = 2 * chebyshev_T(d, x[k] / 2) if d > 0 else 1
   return det(a)

def tquotient(x, la):
   F = x[0].parent()
   return F(tvandermonde(x, la) / vandermonde(x))

def elemz_via_tquotient(z, m):
   n = len(z)
   la = [1] * m + [0] * (n - m)
   return tquotient(z, la)

def test_elemz_various_formulas_numer(n, verbos):
   if verbos >= 2:
      print('test_elemz_various_formulas_numer, n = %d' % n)
   z = random_rational_list(n)
   res0 = elemz_via_lincomb_elem(z)[ : n + 1]
   res1 = [elemz_via_tquotient(z, m) for m in range(n + 1)]
   if verbos >= 3:
      print('res0 = ' + str(res0))
      print('res1 = ' + str(res1))
   return res0 == res1

def big_test_elemz_various_formulas_numer(nmax, nrep, verbos):
   if verbos >= 1:
      print('big_test_elemz_various_formulas_numer, nmax = %d, nrep = %d' % (nmax, nrep))
   tasks = [(n, r) for n in range(1, nmax + 1) for r in range(nrep)]
   return all(test_elemz_various_formulas_numer(n, verbos) for (n, r) in tasks)

def coefT(s, k):
    if (s == 0 and k == 0):
       return 1
    else:
       return (s / (s - k)) * binomial(s - k, k)

def test_elem_via_lincomb_elemz_numer(n, verbos):
   if verbos >= 1:
      print('test_elem_via_lincomb_elemz_numer, n = %d' % n)
   x = random_rational_list_without_zeros(n)
   y = list(x) + [1 / x[j] for j in range(n)]
   z = [x[j] + 1 / x[j] for j in range(n)]
   lhs = elem_pols(z)[ : n + 1]
   elemz_x = elem_pols(y)
   rhs = [0] * (n + 1)
   for m in range(n + 1):
      m2 = floor(m / 2)
      rhs[m] = sum(((-1) ** k) * coefT(n - m + 2 * k, k) * elemz_x[m - 2 * k] for k in range(m2 + 1))
   if verbos >= 3:
      print('lhs = ' + str(lhs))
      print('rhs = ' + str(rhs))
   return lhs == rhs

def big_test_elem_via_lincomb_elemz_numer(nmax, nrep, verbos):
   if verbos >= 1:
      print('big_test_elem_via_lincomb_elemz_polx, nmax = %d, nrep = %d' % (nmax, nrep))
   tasks = [(n, r) for n in range(1, nmax + 1) for r in range(nrep)]
   return all(test_elem_via_lincomb_elemz_numer(n, verbos) for (n, r) in tasks)

def test_elemz_recur_numer(n, verbos):
   if verbos >= 1:
      print('test_elemz_recur_numer, n = %d' % n)
   zext = random_rational_list(n + 1)
   z = zext[ : n]
   F = QQ
   lhs = elemz_via_lincomb_elem(zext)
   elemz_short = elemz_via_lincomb_elem(z)
   rhs0 = elemz_short + [F.zero(), F.zero()]
   rhs1 = [F.zero()] + [zext[n] * elemz_short[j] for j in range(2 * n + 1)] + [F.zero()]
   rhs2 = [F.zero(), F.zero()] + elemz_short
   rhs = [rhs0[j] + rhs1[j] + rhs2[j] for j in range(2 * n + 3)]
   if verbos >= 3:
      print('lhs = ' + str(lhs))
      print('rhs = ' + str(rhs))
   return lhs == rhs

def big_test_elemz_recur_numer(nmax, nrep, verbos):
   if verbos >= 1:
      print('big_test_elemz_recur_numer, nmax = %d, nrep = %d' % (nmax, nrep))
   tasks = [(n, r) for n in range(1, nmax + 1) for r in range(nrep)]
   return all(test_elemz_recur_numer(n, verbos) for (n, r) in tasks)

def test_difference_elemz_numer(n, verbos):
   if verbos >= 1:
      print('test_difference_elemz_numer, n = %d' % n)
   zext12 = random_rational_list(n + 2)
   zshort = zext12[ : n]
   zext1 = zshort + [zext12[n]]
   zext2 = zshort + [zext12[n + 1]]
   F = QQ
   lhs1 = elemz_via_lincomb_elem(zext1)
   lhs2 = elemz_via_lincomb_elem(zext2)
   lhs = [lhs1[m] - lhs2[m] for m in range(2 * n + 3)]
   rhs0 = [F.zero()] + elemz_via_lincomb_elem(zshort) + [F.zero()]
   rhs = [(zext12[n] - zext12[n + 1]) * rhs0[j] for j in range(2 * n + 3)]
   if verbos >= 3:
      print('lhs = ' + str(lhs))
      print('rhs = ' + str(rhs))
   return lhs == rhs

def big_test_difference_elemz_numer(nmax, nrep, verbos):
   if verbos >= 1:
      print('big_test_difference_elemz_numer, nmax = %d, nrep = %d' % (nmax, nrep))
   tasks = [(n, r) for n in range(1, nmax + 1) for r in range(nrep)]
   return all(test_difference_elemz_numer(n, verbos) for (n, r) in tasks)

import time

def all_tests_elemz_numer(nmax, nrep):
   t = time.time()
   verbos = 2
   res = [0] * 5
   res[0] = big_test_elemz_via_lincomb_elem_numer(nmax, nrep, verbos)
   res[1] = big_test_elemz_various_formulas_numer(nmax, nrep, verbos)
   res[2] = big_test_elem_via_lincomb_elemz_numer(nmax, nrep, verbos)
   res[3] = big_test_elemz_recur_numer(nmax, nrep, verbos)
   res[4] = big_test_difference_elemz_numer(nmax, nrep, verbos)
   total_result = all(res)
   t = time.time() - t
   print('time = %.2f seconds' % t)
   print('total result = ' + str(total_result))

# with nmax = 15 and nrep = 20, time < 60 sec
all_tests_elemz_numer(15, 20)

